-module(megaco_text_parser_prev3b).
-define(THIS_MODULE, megaco_text_parser_prev3b).
-export([parse/1, parse_and_scan/1, format_error/1]).

%% The following directive is needed for (significantly) faster compilation
%% of the generated .erl file by the HiPE compiler.  Please do not remove.
-compile([{hipe,[{regalloc,linear_scan}]}]).

-include("megaco_text_parser_prev3b.hrl").



%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, _Endline} ->
            yeccpars1([], false, State, States, Vstack);
        {error, Descriptor, _Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, _Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({_Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(3, __Cat, [0 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(1, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  sep,
 yeccpars2(yeccgoto(optSep, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(2, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 4, [3 | __Ss], [__T | __Stack]);
yeccpars2(3, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(5, __Cat, [3 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(4, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 900, [4 | __Ss], [__T | __Stack]);
yeccpars2(4, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(5, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(8, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(10, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(11, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(12, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(18, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(19, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(20, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(21, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(22, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(24, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(25, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(26, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(27, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(28, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(29, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(49, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(59, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(60, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(71, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(75, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(81, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(83, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(84, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(86, 'endOfMessage', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 899, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(87, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(93, __Cat, [87 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(88, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(89, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 885, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(886, __Cat, [89 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(90, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mId, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(91, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mId, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(92, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 103, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 105, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(93, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'MtpAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(94, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_mtpAddress(__1),
 yeccpars2(yeccgoto(mtpAddress, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(95, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(100, __Cat, [95 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(96, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(99, __Cat, [96 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(97, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {deviceName,__1},
 yeccpars2(yeccgoto(deviceName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(98, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_pathName(__1),
 yeccpars2(yeccgoto(pathName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(99, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(mId, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(100, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(mId, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(101, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 869, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(102, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 881, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(103, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 749, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(104, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 741, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(105, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [105 | __Ss], [__T | __Stack]);
yeccpars2(105, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [105 | __Ss], [__T | __Stack]);
yeccpars2(105, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(106, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {messageError,__1},
 yeccpars2(yeccgoto(messageBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_message(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(message, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(108, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 103, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 105, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(transactionList, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactions,__1},
 yeccpars2(yeccgoto(messageBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(110, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionPending,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionReply,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionRequest,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionResponseAck,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(transactionList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(115, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 731, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(116, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 117, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(117, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(118, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(120, __Cat, [118 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 117, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(120, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(121, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionRequest'{transactionId = asn1_NOVALUE, actions = [__3|__4]},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(transactionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(123, __Cat, [122 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(123, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(actionRequestList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(125, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(126, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_contextID(__1),
 yeccpars2(yeccgoto(contextID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'IEPSToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 128, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'ContextAttrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'EmergencyOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(128, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {addReq,__1},
 yeccpars2(yeccgoto(ammToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(129, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 728, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(130, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 725, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(131, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 709, [131 | __Ss], [__T | __Stack]);
yeccpars2(131, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(132, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 693, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {emergency,false},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {emergency,true},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(135, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 691, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(136, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {modReq,__1},
 yeccpars2(yeccgoto(ammToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(137, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {moveReq,__1},
 yeccpars2(yeccgoto(ammToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 674, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(139, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 672, [139 | __Ss], [__T | __Stack]);
yeccpars2(139, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(140, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 622, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(141, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 616, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(142, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 599, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(143, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 598, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(144, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 594, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(595, __Cat, [144 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(145, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(146, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(147, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(148, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {commandRequest,__1},
 yeccpars2(yeccgoto(actionRequestItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(149, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(150, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {contextAudit,__1},
 yeccpars2(yeccgoto(actionRequestItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(151, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {contextProp,__1},
 yeccpars2(yeccgoto(actionRequestItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {iepsCallind,__1},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(154, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {priority,__1},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(156, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {topology,__1},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(158, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(159, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_terminationID(__1),
 yeccpars2(yeccgoto(terminationID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(160, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(162, __Cat, [160 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(162, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   Descs = merge_AmmRequest_descriptors(__4,[]), make_commandRequest(__1,#'AmmRequest'{terminationID = [__3], descriptors = Descs})
  end,
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(ammRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(163, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 484, [163 | __Ss], [__T | __Stack]);
yeccpars2(163, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(164, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_DMD(__1),
 yeccpars2(yeccgoto(digitMapDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 458, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(eventBufferDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(166, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 395, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'EventsDescriptor'{requestID = asn1_NOVALUE, eventList = []},
 yeccpars2(yeccgoto(eventsDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(167, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(168, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(169, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(170, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 200, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(signalsDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(171, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(172, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(183, __Cat, [172 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(174, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {digitMapDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventBufferDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventsDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mediaDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {modemDescriptor,deprecated},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {muxDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signalsDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(181, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {statisticsDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(183, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [183 | __Ss], [__T | __Stack]);
yeccpars2(183, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(184, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(ammRequestBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(185, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(186, __Cat, [185 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(186, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(ammParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(188, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 196, [188 | __Ss], [__T | __Stack]);
yeccpars2(188, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'StatisticsParameter'{statName = __1, statValue = asn1_NOVALUE},
 yeccpars2(yeccgoto(statisticsParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_pkgdName(__1),
 yeccpars2(yeccgoto(pkgdName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 191, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(192, __Cat, [190 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(192, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 193, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(193, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(statisticsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(194, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 191, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(195, __Cat, [194 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(195, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(statisticsParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(196, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(197, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_value(__1),
 yeccpars2(yeccgoto(value, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(198, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_value(__1),
 yeccpars2(yeccgoto(value, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'StatisticsParameter'{statName = __1, statValue = [__3]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(statisticsParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(200, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(201, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 281, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(202, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(signalName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {seqSigList,__1},
 yeccpars2(yeccgoto(signalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_signalRequest(__1,[]),
 yeccpars2(yeccgoto(signalRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(208, __Cat, [205 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(206, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signal,__1},
 yeccpars2(yeccgoto(signalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(207, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(208, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(209, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(signalsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(211, __Cat, [210 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(211, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(signalParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(212, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 219, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 218, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 216, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 214, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(213, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 276, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 274, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, 'COMMA', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(sigParameter, hd(__Ss)), 'COMMA', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, 'RBRKT', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(sigParameter, hd(__Ss)), 'RBRKT', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 262, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 259, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(218, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 254, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(219, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 230, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 227, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(221, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 222, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(223, __Cat, [221 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(222, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 219, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 218, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 216, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 214, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(223, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 224, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(224, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_signalRequest(__1,[__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(signalRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 222, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(226, __Cat, [225 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(226, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 236, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 235, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(228, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(229, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(230, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(231, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {other,ensure_NAME(__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(232, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2], extraInfo = {relation,unequalTo}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(233, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2], extraInfo = {relation,smallerThan}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2], extraInfo = {relation,greaterThan}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(236, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(237, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(238, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__1]},
 yeccpars2(yeccgoto(alternativeValue, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(239, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 241, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(242, __Cat, [239 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(240, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(241, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(242, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 243, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(243, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2|__3], extraInfo = {sublist,true}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(alternativeValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(244, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 241, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(245, __Cat, [244 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(245, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(valueList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(246, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(247, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2,__4], extraInfo = {range,true}},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(alternativeValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(248, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 241, [248 | __Ss], [__T | __Stack]);
yeccpars2(248, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(249, __Cat, [248 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(249, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 250, [249 | __Ss], [__T | __Stack]);
yeccpars2(249, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(250, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2|__3], extraInfo = {sublist,false}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(alternativeValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(251, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(252, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_streamID(__1),
 yeccpars2(yeccgoto(streamID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(253, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {stream,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 255, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 257, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(255, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  brief,
 yeccpars2(yeccgoto(signalType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(256, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onOff,
 yeccpars2(yeccgoto(signalType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(257, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  timeOut,
 yeccpars2(yeccgoto(signalType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(258, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signal_type,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(259, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(260, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {requestId,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(261, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_requestID(__1),
 yeccpars2(yeccgoto(requestID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(262, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 263, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(263, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 266, [263 | __Ss], [__T | __Stack]);
yeccpars2(263, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 265, [263 | __Ss], [__T | __Stack]);
yeccpars2(263, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [263 | __Ss], [__T | __Stack]);
yeccpars2(263, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 267, [263 | __Ss], [__T | __Stack]);
yeccpars2(263, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(264, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onInterruptByEvent,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(265, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onInterruptByNewSignalDescr,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(266, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  otherReason,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onTimeOut,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(268, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(270, __Cat, [268 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(269, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 266, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 265, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 267, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(270, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 271, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(271, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {notify_completion,[__4|__5]},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(272, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(273, __Cat, [272 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(273, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(notificationReasons, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(274, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(275, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {duration,ensure_uint16(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(276, 'BothToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 277, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, 'InternalToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 279, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, 'ExternalToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 278, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(277, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  both,
 yeccpars2(yeccgoto(direction, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(278, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  external,
 yeccpars2(yeccgoto(direction, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(279, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  internal,
 yeccpars2(yeccgoto(direction, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(280, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {direction,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(281, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(282, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint16(__1),
 yeccpars2(yeccgoto(signalListId, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(283, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(284, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(285, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 287, [285 | __Ss], [__T | __Stack]);
yeccpars2(285, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(288, __Cat, [285 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(286, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(signalListParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(287, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(288, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(289, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SeqSigList'{id = ensure_uint16(__3), signalList = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(signalList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(290, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 287, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(291, __Cat, [290 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(291, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(signalListParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(292, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(293, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 295, [293 | __Ss], [__T | __Stack]);
yeccpars2(293, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(294, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_muxType(__1),
 yeccpars2(yeccgoto(muxType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(295, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(296, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'MuxDescriptor'{muxType = __3, termList = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(muxDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(297, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(299, __Cat, [297 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(298, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(299, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 300, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(300, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(terminationIDList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(301, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(302, __Cat, [301 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(302, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(terminationIDListRepeat, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(303, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(304, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(305, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 yeccpars2(308, __Cat, [305 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(306, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 yeccpars2(yeccgoto(modemType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(307, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(308, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 309, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(309, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 310, [309 | __Ss], [__T | __Stack]);
yeccpars2(309, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(311, __Cat, [309 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(310, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(311, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(modemDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(312, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 230, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 227, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(313, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [313 | __Ss], [__T | __Stack]);
yeccpars2(313, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(315, __Cat, [313 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(314, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(315, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 316, [315 | __Ss], [__T | __Stack]);
yeccpars2(315, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(316, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(optPropertyParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(317, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(318, __Cat, [317 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(318, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(propertyParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(319, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'PropertyParm'.name,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(propertyParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(320, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [320 | __Ss], [__T | __Stack]);
yeccpars2(320, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 yeccpars2(321, __Cat, [320 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(321, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(modemTypeList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(322, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 310, [322 | __Ss], [__T | __Stack]);
yeccpars2(322, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(323, __Cat, [322 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(323, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(modemDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(324, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 329, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 328, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 327, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 326, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(325, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 371, [325 | __Ss], [__T | __Stack]);
yeccpars2(325, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(326, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {local,#'LocalRemoteDescriptor'{propGrps = ensure_prop_groups(__1)}},
 yeccpars2(yeccgoto(streamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(327, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {remote,#'LocalRemoteDescriptor'{propGrps = ensure_prop_groups(__1)}},
 yeccpars2(yeccgoto(streamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(328, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 362, [328 | __Ss], [__T | __Stack]);
yeccpars2(328, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(329, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 341, [329 | __Ss], [__T | __Stack]);
yeccpars2(329, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(330, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {control,__1},
 yeccpars2(yeccgoto(streamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(331, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 336, [331 | __Ss], [__T | __Stack]);
yeccpars2(331, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(337, __Cat, [331 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(332, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {statistics,__1},
 yeccpars2(yeccgoto(streamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(333, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamDescriptor,__1},
 yeccpars2(yeccgoto(mediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(334, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamParm,__1},
 yeccpars2(yeccgoto(mediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(335, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {termState,__1},
 yeccpars2(yeccgoto(mediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(336, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 329, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 328, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 327, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 326, [336 | __Ss], [__T | __Stack]);
yeccpars2(336, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(337, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 338, [337 | __Ss], [__T | __Stack]);
yeccpars2(337, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(338, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_mediaDescriptor([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(mediaDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(339, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 336, [339 | __Ss], [__T | __Stack]);
yeccpars2(339, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(340, __Cat, [339 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(340, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(mediaParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(341, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 343, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 342, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(342, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 358, [342 | __Ss], [__T | __Stack]);
yeccpars2(342, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(343, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 353, [343 | __Ss], [__T | __Stack]);
yeccpars2(343, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(344, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventBufferControl,__1},
 yeccpars2(yeccgoto(terminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(345, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {propertyParm,__1},
 yeccpars2(yeccgoto(terminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(346, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {serviceState,__1},
 yeccpars2(yeccgoto(terminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(347, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 348, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(349, __Cat, [347 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(348, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 343, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 342, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(349, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 350, [349 | __Ss], [__T | __Stack]);
yeccpars2(349, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(350, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_terminationStateDescriptor([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(terminationStateDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(351, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 348, [351 | __Ss], [__T | __Stack]);
yeccpars2(351, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(352, __Cat, [351 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(352, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(terminationStateParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(353, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 354, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 355, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 356, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(354, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  inSvc,
 yeccpars2(yeccgoto(serviceState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(355, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  outOfSvc,
 yeccpars2(yeccgoto(serviceState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(356, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  test,
 yeccpars2(yeccgoto(serviceState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(357, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceStates, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(358, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 359, [358 | __Ss], [__T | __Stack]);
yeccpars2(358, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 360, [358 | __Ss], [__T | __Stack]);
yeccpars2(358, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(359, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  lockStep,
 yeccpars2(yeccgoto(eventBufferControlState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(360, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  off,
 yeccpars2(yeccgoto(eventBufferControlState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(361, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventBufferControl, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(362, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(363, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 364, [363 | __Ss], [__T | __Stack]);
yeccpars2(363, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(364, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 327, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 326, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(365, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 366, [365 | __Ss], [__T | __Stack]);
yeccpars2(365, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(367, __Cat, [365 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(366, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [366 | __Ss], [__T | __Stack]);
yeccpars2(366, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [366 | __Ss], [__T | __Stack]);
yeccpars2(366, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 327, [366 | __Ss], [__T | __Stack]);
yeccpars2(366, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 326, [366 | __Ss], [__T | __Stack]);
yeccpars2(366, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(367, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 368, [367 | __Ss], [__T | __Stack]);
yeccpars2(367, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(368, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'StreamDescriptor'{streamID = __3, streamParms = merge_streamParms([__5|__6])},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(streamDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(369, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 366, [369 | __Ss], [__T | __Stack]);
yeccpars2(369, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(370, __Cat, [369 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(370, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(streamParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(371, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 374, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 373, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 372, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(372, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 388, [372 | __Ss], [__T | __Stack]);
yeccpars2(372, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(373, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 386, [373 | __Ss], [__T | __Stack]);
yeccpars2(373, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(374, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 382, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(375, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 377, [375 | __Ss], [__T | __Stack]);
yeccpars2(375, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(378, __Cat, [375 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(376, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {prop,__1},
 yeccpars2(yeccgoto(localParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(377, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 374, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 373, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 372, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(378, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 379, [378 | __Ss], [__T | __Stack]);
yeccpars2(378, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(379, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(localControlDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(380, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 377, [380 | __Ss], [__T | __Stack]);
yeccpars2(380, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(381, __Cat, [380 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(381, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(382, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 383, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 384, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(383, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  false,
 yeccpars2(yeccgoto(onOrOff, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(384, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  true,
 yeccpars2(yeccgoto(onOrOff, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(385, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {value,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(386, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 383, [386 | __Ss], [__T | __Stack]);
yeccpars2(386, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 384, [386 | __Ss], [__T | __Stack]);
yeccpars2(386, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(387, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {group,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(388, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 390, [388 | __Ss], [__T | __Stack]);
yeccpars2(388, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 389, [388 | __Ss], [__T | __Stack]);
yeccpars2(388, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 393, [388 | __Ss], [__T | __Stack]);
yeccpars2(388, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 391, [388 | __Ss], [__T | __Stack]);
yeccpars2(388, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 392, [388 | __Ss], [__T | __Stack]);
yeccpars2(388, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(389, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  inactive,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(390, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  loopBack,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(391, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  recvOnly,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(392, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  sendOnly,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(393, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  sendRecv,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(394, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mode,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(395, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(396, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 397, [396 | __Ss], [__T | __Stack]);
yeccpars2(396, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(397, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [397 | __Ss], [__T | __Stack]);
yeccpars2(397, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(398, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 405, [398 | __Ss], [__T | __Stack]);
yeccpars2(398, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  #'RequestedEvent'{evParList = []},
 yeccpars2(406, __Cat, [398 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(399, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 400, [399 | __Ss], [__T | __Stack]);
yeccpars2(399, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(401, __Cat, [399 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(400, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(401, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 402, [401 | __Ss], [__T | __Stack]);
yeccpars2(401, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(402, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'EventsDescriptor'{requestID = __3, eventList = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(eventsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(403, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 400, [403 | __Ss], [__T | __Stack]);
yeccpars2(403, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(404, __Cat, [403 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(404, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(requestedEvents, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(405, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 409, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 408, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 407, [405 | __Ss], [__T | __Stack]);
yeccpars2(405, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(406, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'RequestedEvent'.pkgdName,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(requestedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(407, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_eventDM(__1),
 yeccpars2(yeccgoto(eventDM, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(408, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 423, [408 | __Ss], [__T | __Stack]);
yeccpars2(408, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(409, 'COMMA', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), 'COMMA', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(409, 'RBRKT', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), 'RBRKT', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(409, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(410, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(411, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(412, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(413, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 418, [413 | __Ss], [__T | __Stack]);
yeccpars2(413, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(419, __Cat, [413 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(414, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 230, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 227, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(415, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(416, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_NAME(__1),
 yeccpars2(yeccgoto(eventParameterName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(417, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  select_stream_or_other(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(eventStreamOrOther, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(418, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 409, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 408, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 407, [418 | __Ss], [__T | __Stack]);
yeccpars2(418, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(419, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 420, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(420, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_eventParameters([__2|__3]),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(requestedEventBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(421, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 418, [421 | __Ss], [__T | __Stack]);
yeccpars2(421, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(422, __Cat, [421 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(422, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(423, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [423 | __Ss], [__T | __Stack]);
yeccpars2(423, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 424, [423 | __Ss], [__T | __Stack]);
yeccpars2(423, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(424, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 432, [424 | __Ss], [__T | __Stack]);
yeccpars2(424, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SecondEventsDescriptor'{requestID = asn1_NOVALUE, eventList = []},
 yeccpars2(yeccgoto(embedFirst, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(425, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 431, [425 | __Ss], [__T | __Stack]);
yeccpars2(425, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(426, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 428, [426 | __Ss], [__T | __Stack]);
yeccpars2(426, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 427, [426 | __Ss], [__T | __Stack]);
yeccpars2(426, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(427, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 424, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(428, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {embed,__3,asn1_NOVALUE},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(embedWithSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(429, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 430, [429 | __Ss], [__T | __Stack]);
yeccpars2(429, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(430, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {embed,__3,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(embedWithSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(431, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {embed,asn1_NOVALUE,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(embedNoSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(432, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(433, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 434, [433 | __Ss], [__T | __Stack]);
yeccpars2(433, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(434, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [434 | __Ss], [__T | __Stack]);
yeccpars2(434, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(435, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 442, [435 | __Ss], [__T | __Stack]);
yeccpars2(435, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  #'SecondRequestedEvent'{evParList = []},
 yeccpars2(443, __Cat, [435 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(436, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 437, [436 | __Ss], [__T | __Stack]);
yeccpars2(436, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(438, __Cat, [436 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(437, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(438, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 439, [438 | __Ss], [__T | __Stack]);
yeccpars2(438, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(439, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SecondEventsDescriptor'{requestID = __3, eventList = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(embedFirst, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(440, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 437, [440 | __Ss], [__T | __Stack]);
yeccpars2(440, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(441, __Cat, [440 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(441, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(secondRequestedEvents, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(442, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 445, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 444, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 407, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(443, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'SecondRequestedEvent'.pkgdName,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(secondRequestedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(444, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 455, [444 | __Ss], [__T | __Stack]);
yeccpars2(444, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(445, 'COMMA', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), 'COMMA', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(445, 'RBRKT', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), 'RBRKT', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(445, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(446, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(447, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(448, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(449, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 450, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(451, __Cat, [449 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(450, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 445, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 444, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 407, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(451, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 452, [451 | __Ss], [__T | __Stack]);
yeccpars2(451, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(452, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_secondEventParameters([__2|__3]),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(secondRequestedEventBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(453, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 450, [453 | __Ss], [__T | __Stack]);
yeccpars2(453, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(454, __Cat, [453 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(454, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(secondEventParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(455, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [455 | __Ss], [__T | __Stack]);
yeccpars2(455, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(456, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 457, [456 | __Ss], [__T | __Stack]);
yeccpars2(456, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(457, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {second_embed,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(embedSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(458, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [458 | __Ss], [__T | __Stack]);
yeccpars2(458, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [458 | __Ss], [__T | __Stack]);
yeccpars2(458, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(462, __Cat, [458 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(459, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_timeStamp(__1),
 yeccpars2(yeccgoto(timeStamp, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(460, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 479, [460 | __Ss], [__T | __Stack]);
yeccpars2(460, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(480, __Cat, [460 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(461, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_eventSpec(__1),
 yeccpars2(yeccgoto(eventSpec, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(462, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [462 | __Ss], [__T | __Stack]);
yeccpars2(462, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(463, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [463 | __Ss], [__T | __Stack]);
yeccpars2(463, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(464, __Cat, [463 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(464, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 465, [464 | __Ss], [__T | __Stack]);
yeccpars2(464, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(465, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [465 | __Ss], [__T | __Stack]);
yeccpars2(465, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(466, __Cat, [465 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(466, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(467, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 468, [467 | __Ss], [__T | __Stack]);
yeccpars2(467, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(469, __Cat, [467 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(468, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [468 | __Ss], [__T | __Stack]);
yeccpars2(468, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(469, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_observed_event(__6,__5,__1),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(observedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(470, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(observedEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(471, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 472, [471 | __Ss], [__T | __Stack]);
yeccpars2(471, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(473, __Cat, [471 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(472, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(473, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 474, [473 | __Ss], [__T | __Stack]);
yeccpars2(473, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(474, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(observedEventBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(475, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 472, [475 | __Ss], [__T | __Stack]);
yeccpars2(475, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(476, __Cat, [475 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(476, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(observedEventParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(477, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 468, [477 | __Ss], [__T | __Stack]);
yeccpars2(477, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(478, __Cat, [477 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(478, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_observed_event(__3,__2,asn1_NOVALUE),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(observedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(479, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [479 | __Ss], [__T | __Stack]);
yeccpars2(479, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [479 | __Ss], [__T | __Stack]);
yeccpars2(479, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(462, __Cat, [479 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(480, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 481, [480 | __Ss], [__T | __Stack]);
yeccpars2(480, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(481, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(eventBufferDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(482, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 479, [482 | __Ss], [__T | __Stack]);
yeccpars2(482, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(483, __Cat, [482 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(483, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventSpecList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(484, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 493, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 495, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 485, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 494, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 488, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 487, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 489, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 492, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 486, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 490, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 491, [484 | __Ss], [__T | __Stack]);
yeccpars2(484, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(496, __Cat, [484 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(485, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_IADMD(__1),
 yeccpars2(yeccgoto(indAuddigitMapDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(486, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  digitMapToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(487, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 576, [487 | __Ss], [__T | __Stack]);
yeccpars2(487, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  eventBufferToken,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(488, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 570, [488 | __Ss], [__T | __Stack]);
yeccpars2(488, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  eventsToken,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(489, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 538, [489 | __Ss], [__T | __Stack]);
yeccpars2(489, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mediaToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(490, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  modemToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(491, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  muxToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(492, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  observedEventsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(493, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 534, [493 | __Ss], [__T | __Stack]);
yeccpars2(493, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  packagesToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(494, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 521, [494 | __Ss], [__T | __Stack]);
yeccpars2(494, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  signalsToken,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(495, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 518, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  statsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(496, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 593, [496 | __Ss], [__T | __Stack]);
yeccpars2(496, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(497, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 589, [497 | __Ss], [__T | __Stack]);
yeccpars2(497, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(590, __Cat, [497 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(498, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(499, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 508, [499 | __Ss], [__T | __Stack]);
yeccpars2(499, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(509, __Cat, [499 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(500, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudDigitMapDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(501, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudEventBufferDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(502, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudEventsDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(503, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudMediaDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(504, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudPackagesDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(505, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudSignalsDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(506, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudStatisticsDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(507, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {terminationAudit,__1},
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(508, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 513, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 515, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 485, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 514, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 511, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 510, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 512, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(509, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(indAudterminationAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(510, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 576, [510 | __Ss], [__T | __Stack]);
yeccpars2(510, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(511, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 570, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(512, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 538, [512 | __Ss], [__T | __Stack]);
yeccpars2(512, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(513, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 534, [513 | __Ss], [__T | __Stack]);
yeccpars2(513, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(514, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 521, [514 | __Ss], [__T | __Stack]);
yeccpars2(514, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(515, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 518, [515 | __Ss], [__T | __Stack]);
yeccpars2(515, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(516, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 508, [516 | __Ss], [__T | __Stack]);
yeccpars2(516, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(517, __Cat, [516 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(517, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(indAudterminationAuditList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(518, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(519, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 520, [519 | __Ss], [__T | __Stack]);
yeccpars2(519, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(520, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudStatisticsDescriptor'{statName = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudstatisticsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(521, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 524, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 523, [521 | __Ss], [__T | __Stack]);
yeccpars2(521, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(522, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(indAudsignalsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(523, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(optIndAudsignalParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(524, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 529, [524 | __Ss], [__T | __Stack]);
yeccpars2(524, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(525, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {seqSigList,__1},
 yeccpars2(yeccgoto(indAudsignalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(526, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 528, [526 | __Ss], [__T | __Stack]);
yeccpars2(526, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(527, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signal,ensure_indAudSignal(__1)},
 yeccpars2(yeccgoto(indAudsignalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(528, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(optIndAudsignalParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(529, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(530, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 531, [530 | __Ss], [__T | __Stack]);
yeccpars2(530, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(531, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(532, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 533, [532 | __Ss], [__T | __Stack]);
yeccpars2(532, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(533, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudSeqSigList'{id = ensure_uint16(__3), signalList = ensure_indAudSignalListParm(__5)},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(indAudsignalList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(534, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(535, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 537, [535 | __Ss], [__T | __Stack]);
yeccpars2(535, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(536, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_packagesItem(__1),
 yeccpars2(yeccgoto(packagesItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(537, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudPackagesDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudpackagesDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(538, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 515, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 541, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 539, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 540, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(539, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 562, [539 | __Ss], [__T | __Stack]);
yeccpars2(539, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(540, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 557, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(541, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 553, [541 | __Ss], [__T | __Stack]);
yeccpars2(541, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(542, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudStreamParms'{localControlDescriptor = __1},
 yeccpars2(yeccgoto(indAudstreamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(543, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 548, [543 | __Ss], [__T | __Stack]);
yeccpars2(543, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(549, __Cat, [543 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(544, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudStreamParms'{statisticsDescriptor = __1},
 yeccpars2(yeccgoto(indAudstreamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(545, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamDescr,__1},
 yeccpars2(yeccgoto(indAudmediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(546, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamParm,__1},
 yeccpars2(yeccgoto(indAudmediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(547, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {termStateDescr,__1},
 yeccpars2(yeccgoto(indAudmediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(548, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 515, [548 | __Ss], [__T | __Stack]);
yeccpars2(548, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 541, [548 | __Ss], [__T | __Stack]);
yeccpars2(548, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 539, [548 | __Ss], [__T | __Stack]);
yeccpars2(548, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 540, [548 | __Ss], [__T | __Stack]);
yeccpars2(548, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(549, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 550, [549 | __Ss], [__T | __Stack]);
yeccpars2(549, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(550, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudMediaDescriptor([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(indAudmediaDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(551, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 548, [551 | __Ss], [__T | __Stack]);
yeccpars2(551, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(552, __Cat, [551 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(552, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(indAudmediaParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(553, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [553 | __Ss], [__T | __Stack]);
yeccpars2(553, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(554, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 556, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(555, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_indAudTerminationStateParm(__1),
 yeccpars2(yeccgoto(indAudterminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(556, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudTerminationStateDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudterminationStateDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(557, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(558, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 559, [558 | __Ss], [__T | __Stack]);
yeccpars2(558, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(559, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 515, [559 | __Ss], [__T | __Stack]);
yeccpars2(559, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 539, [559 | __Ss], [__T | __Stack]);
yeccpars2(559, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(560, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 561, [560 | __Ss], [__T | __Stack]);
yeccpars2(560, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(561, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudStreamDescriptor'{streamID = __3, streamParms = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(indAudstreamDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(562, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(563, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 565, [563 | __Ss], [__T | __Stack]);
yeccpars2(563, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(566, __Cat, [563 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(564, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_indAudLocalParm(__1),
 yeccpars2(yeccgoto(indAudlocalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(565, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(566, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 567, [566 | __Ss], [__T | __Stack]);
yeccpars2(566, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(567, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudLocalControlDescriptor([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(indAudlocalControlDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(568, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 565, [568 | __Ss], [__T | __Stack]);
yeccpars2(568, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(569, __Cat, [568 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(569, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(indAudlocalParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(570, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(571, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 572, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(572, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(573, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 575, [573 | __Ss], [__T | __Stack]);
yeccpars2(573, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(574, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(indAudrequestedEvent, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(575, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudEventsDescriptor'{requestID = __3, pkgdName = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(indAudeventsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(576, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(577, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 588, [577 | __Ss], [__T | __Stack]);
yeccpars2(577, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(578, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 579, [578 | __Ss], [__T | __Stack]);
yeccpars2(578, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(580, __Cat, [578 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(579, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 581, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(580, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudEventBufferDescriptor(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(indAudeventSpec, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(581, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 586, [581 | __Ss], [__T | __Stack]);
yeccpars2(581, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(582, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventParameterName,__1},
 yeccpars2(yeccgoto(indAudeventSpecParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(583, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamID,__1},
 yeccpars2(yeccgoto(indAudeventSpecParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(584, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 585, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(585, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(optIndAudeventSpecParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(586, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(587, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventStream, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(588, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudeventBufferDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(589, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 493, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 495, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 485, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 494, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 488, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 487, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 489, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 492, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 486, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 490, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 491, [589 | __Ss], [__T | __Stack]);
yeccpars2(589, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(590, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(auditDescriptorBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(591, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 589, [591 | __Ss], [__T | __Stack]);
yeccpars2(591, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(592, __Cat, [591 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(592, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditItemList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(593, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_auditDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(594, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'IEPSToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 128, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'ContextAttrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'EmergencyOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(595, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(actionRequestBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(596, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 594, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(597, __Cat, [596 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(597, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(actionRequestItems, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(598, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_action_request(__3,__5),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(actionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(599, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(600, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 608, [600 | __Ss], [__T | __Stack]);
yeccpars2(600, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(601, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(terminationA, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(602, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 603, [602 | __Ss], [__T | __Stack]);
yeccpars2(602, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(604, __Cat, [602 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(603, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(604, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 605, [604 | __Ss], [__T | __Stack]);
yeccpars2(604, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(605, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(topologyDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(606, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 603, [606 | __Ss], [__T | __Stack]);
yeccpars2(606, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(607, __Cat, [606 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(607, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(topologyTripleList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(608, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [608 | __Ss], [__T | __Stack]);
yeccpars2(608, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(609, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 611, [609 | __Ss], [__T | __Stack]);
yeccpars2(609, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(610, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(terminationB, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(611, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 614, [611 | __Ss], [__T | __Stack]);
yeccpars2(611, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 613, [611 | __Ss], [__T | __Stack]);
yeccpars2(611, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 612, [611 | __Ss], [__T | __Stack]);
yeccpars2(611, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(612, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  bothway,
 yeccpars2(yeccgoto(topologyDirection, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(613, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  isolate,
 yeccpars2(yeccgoto(topologyDirection, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(614, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  oneway,
 yeccpars2(yeccgoto(topologyDirection, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(615, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TopologyRequest'{terminationFrom = __1, terminationTo = __3, topologyDirection = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(topologyTriple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(616, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [616 | __Ss], [__T | __Stack]);
yeccpars2(616, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(617, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 618, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(619, __Cat, [617 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(618, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [618 | __Ss], [__T | __Stack]);
yeccpars2(618, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(619, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({subtractReq,__1},#'SubtractRequest'{terminationID = [__3], auditDescriptor = __4}),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(subtractRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(620, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 621, [620 | __Ss], [__T | __Stack]);
yeccpars2(620, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(621, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(optAuditDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(622, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [622 | __Ss], [__T | __Stack]);
yeccpars2(622, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(623, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 624, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(624, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 625, [624 | __Ss], [__T | __Stack]);
yeccpars2(624, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(625, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 628, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(626, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 627, [626 | __Ss], [__T | __Stack]);
yeccpars2(626, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(627, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({serviceChangeReq,__1},#'ServiceChangeRequest'{terminationID = [__3], serviceChangeParms = __5}),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(serviceChangeRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(628, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 636, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 634, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 633, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 632, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 631, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 630, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 629, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ServiceChangeIncompleteToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 635, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 493, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 495, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 485, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 494, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 488, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 487, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 489, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 492, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 486, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 490, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 491, [628 | __Ss], [__T | __Stack]);
yeccpars2(628, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(629, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 670, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(630, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 668, [630 | __Ss], [__T | __Stack]);
yeccpars2(630, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(631, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 666, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(632, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 664, [632 | __Ss], [__T | __Stack]);
yeccpars2(632, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(633, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 662, [633 | __Ss], [__T | __Stack]);
yeccpars2(633, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(634, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 658, [634 | __Ss], [__T | __Stack]);
yeccpars2(634, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(635, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  incomplete,
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(636, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 656, [636 | __Ss], [__T | __Stack]);
yeccpars2(636, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(637, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {audit_item,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(638, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {extension,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(639, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 230, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 227, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(640, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_extensionParameter(__1),
 yeccpars2(yeccgoto(extensionParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(641, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {address,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(642, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {delay,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(643, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {method,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(644, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mgc_id,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(645, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 650, [645 | __Ss], [__T | __Stack]);
yeccpars2(645, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(651, __Cat, [645 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(646, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {profile,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(647, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {reason,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(648, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {version,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(649, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {time_stamp,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(650, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 636, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 634, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 633, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 632, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 631, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 630, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 629, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ServiceChangeIncompleteToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 635, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 493, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 495, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 485, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 494, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 488, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 487, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 489, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 492, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 486, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 490, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 491, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(651, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 652, [651 | __Ss], [__T | __Stack]);
yeccpars2(651, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(652, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_ServiceChangeParm([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(serviceChangeDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(653, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 650, [653 | __Ss], [__T | __Stack]);
yeccpars2(653, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(654, __Cat, [653 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(654, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(655, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'PropertyParm'.name,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(extension, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(656, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(657, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_version(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeVersion, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(658, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(93, __Cat, [658 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(659, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(660, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {portNumber,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(661, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint16(__1),
 yeccpars2(yeccgoto(portNumber, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(662, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(663, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeReason, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(664, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(665, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_profile(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeProfile, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(666, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [666 | __Ss], [__T | __Stack]);
yeccpars2(666, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [666 | __Ss], [__T | __Stack]);
yeccpars2(666, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [666 | __Ss], [__T | __Stack]);
yeccpars2(666, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(93, __Cat, [666 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(667, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeMgcId, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(668, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(669, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_serviceChangeMethod(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeMethod, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(670, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [670 | __Ss], [__T | __Stack]);
yeccpars2(670, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(671, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint32(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeDelay, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(672, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(673, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint16(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(priority, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(674, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(675, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 676, [675 | __Ss], [__T | __Stack]);
yeccpars2(675, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(676, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 677, [676 | __Ss], [__T | __Stack]);
yeccpars2(676, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [676 | __Ss], [__T | __Stack]);
yeccpars2(676, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(677, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 682, [677 | __Ss], [__T | __Stack]);
yeccpars2(677, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(678, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'NotifyRequest'{errorDescriptor = __1},
 yeccpars2(yeccgoto(notifyRequestBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(679, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 681, [679 | __Ss], [__T | __Stack]);
yeccpars2(679, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(680, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'NotifyRequest'{observedEventsDescriptor = __1},
 yeccpars2(yeccgoto(notifyRequestBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(681, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({notifyReq,__1},setelement(#'NotifyRequest'.terminationID,__5,[__3])),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(notifyRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(682, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(683, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 684, [683 | __Ss], [__T | __Stack]);
yeccpars2(683, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(684, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(462, __Cat, [684 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(685, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 686, [685 | __Ss], [__T | __Stack]);
yeccpars2(685, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(687, __Cat, [685 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(686, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [686 | __Ss], [__T | __Stack]);
yeccpars2(686, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [686 | __Ss], [__T | __Stack]);
yeccpars2(686, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(462, __Cat, [686 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(687, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 688, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(688, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ObservedEventsDescriptor'{requestId = __3, observedEventLst = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(observedEventsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(689, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 686, [689 | __Ss], [__T | __Stack]);
yeccpars2(689, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(690, __Cat, [689 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(690, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(observedEvents, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(691, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 383, [691 | __Ss], [__T | __Stack]);
yeccpars2(691, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 384, [691 | __Ss], [__T | __Stack]);
yeccpars2(691, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(692, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(iepsValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(693, 'ContextAttrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 694, [693 | __Ss], [__T | __Stack]);
yeccpars2(693, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(694, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 697, [694 | __Ss], [__T | __Stack]);
yeccpars2(694, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(695, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 696, [695 | __Ss], [__T | __Stack]);
yeccpars2(695, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(696, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_context_attr_audit_request(#'ContextAttrAuditRequest'{},__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(contextAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(697, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'IEPSToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 699, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 700, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 698, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 701, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(698, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  emergencyAudit,
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(699, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  iepsCallind,
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(700, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  priorityAudit,
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(701, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  topologyAudit,
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(702, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 704, [702 | __Ss], [__T | __Stack]);
yeccpars2(702, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(705, __Cat, [702 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(703, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {prop,__1},
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(704, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'IEPSToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 699, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 700, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 698, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 701, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(705, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 706, [705 | __Ss], [__T | __Stack]);
yeccpars2(705, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(706, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(indAudcontextAttrDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(707, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 704, [707 | __Ss], [__T | __Stack]);
yeccpars2(707, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(708, __Cat, [707 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(708, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(contextAuditProperties, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(709, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, 'ContextListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 710, [709 | __Ss], [__T | __Stack]);
yeccpars2(709, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(710, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 717, [710 | __Ss], [__T | __Stack]);
yeccpars2(710, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(711, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 716, [711 | __Ss], [__T | __Stack]);
yeccpars2(711, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(712, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [712 | __Ss], [__T | __Stack]);
yeccpars2(712, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(715, __Cat, [712 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(713, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 714, [713 | __Ss], [__T | __Stack]);
yeccpars2(713, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(714, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {contextProp,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(contextAttrDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(715, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(propertyParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(716, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {contextList,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(contextAttrDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(717, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 718, [717 | __Ss], [__T | __Stack]);
yeccpars2(717, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(718, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [718 | __Ss], [__T | __Stack]);
yeccpars2(718, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(719, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 720, [719 | __Ss], [__T | __Stack]);
yeccpars2(719, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(721, __Cat, [719 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(720, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [720 | __Ss], [__T | __Stack]);
yeccpars2(720, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(721, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 722, [721 | __Ss], [__T | __Stack]);
yeccpars2(721, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(722, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__4|__5],
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(contextIdList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(723, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 720, [723 | __Ss], [__T | __Stack]);
yeccpars2(723, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(724, __Cat, [723 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(724, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(contextIDs, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(725, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [725 | __Ss], [__T | __Stack]);
yeccpars2(725, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(726, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 618, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(727, __Cat, [726 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(727, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({auditValueRequest,__1},#'AuditRequest'{terminationID = __3, auditDescriptor = __4}),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(728, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(729, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 618, [729 | __Ss], [__T | __Stack]);
yeccpars2(729, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(730, __Cat, [729 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(730, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({auditCapRequest,__1},#'AuditRequest'{terminationID = __3, auditDescriptor = __4}),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(731, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 117, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(732, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint32(__1),
 yeccpars2(yeccgoto(transactionID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(733, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 734, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(734, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 117, [734 | __Ss], [__T | __Stack]);
yeccpars2(734, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(735, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [735 | __Ss], [__T | __Stack]);
yeccpars2(735, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(736, __Cat, [735 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(736, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 737, [736 | __Ss], [__T | __Stack]);
yeccpars2(736, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(737, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionRequest'{transactionId = ensure_transactionID(__3), actions = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(transactionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(738, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [738 | __Ss], [__T | __Stack]);
yeccpars2(738, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(739, __Cat, [738 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(739, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 740, [739 | __Ss], [__T | __Stack]);
yeccpars2(739, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(740, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionRequest'{transactionId = asn1_NOVALUE, actions = [__4|__5]},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(transactionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(741, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [741 | __Ss], [__T | __Stack]);
yeccpars2(741, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(742, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_transactionAck(__1),
 yeccpars2(yeccgoto(transactionAck, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(743, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 744, [743 | __Ss], [__T | __Stack]);
yeccpars2(743, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(745, __Cat, [743 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(744, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [744 | __Ss], [__T | __Stack]);
yeccpars2(744, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(745, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 746, [745 | __Ss], [__T | __Stack]);
yeccpars2(745, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(746, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(transactionResponseAck, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(747, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 744, [747 | __Ss], [__T | __Stack]);
yeccpars2(747, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(748, __Cat, [747 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(748, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(transactionAckList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(749, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [749 | __Ss], [__T | __Stack]);
yeccpars2(749, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(750, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 751, [750 | __Ss], [__T | __Stack]);
yeccpars2(750, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(751, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 752, [751 | __Ss], [__T | __Stack]);
yeccpars2(751, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(753, __Cat, [751 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(752, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 880, [752 | __Ss], [__T | __Stack]);
yeccpars2(752, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(753, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [753 | __Ss], [__T | __Stack]);
yeccpars2(753, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 754, [753 | __Ss], [__T | __Stack]);
yeccpars2(753, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(754, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 763, [754 | __Ss], [__T | __Stack]);
yeccpars2(754, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(755, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 759, [755 | __Ss], [__T | __Stack]);
yeccpars2(755, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(760, __Cat, [755 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(756, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionError,__1},
 yeccpars2(yeccgoto(transactionReplyBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(757, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 758, [757 | __Ss], [__T | __Stack]);
yeccpars2(757, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(758, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionReply'{transactionId = __3, immAckRequired = __5, transactionResult = __6},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(transactionReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(759, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 754, [759 | __Ss], [__T | __Stack]);
yeccpars2(759, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(760, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {actionReplies,[__1|__2]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(transactionReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(761, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 759, [761 | __Ss], [__T | __Stack]);
yeccpars2(761, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(762, __Cat, [761 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(762, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(actionReplyList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(763, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(764, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 765, [764 | __Ss], [__T | __Stack]);
yeccpars2(764, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(765, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'IEPSToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 772, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 771, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 767, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 768, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 773, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 769, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 770, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 766, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'ContextAttrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'EmergencyOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [765 | __Ss], [__T | __Stack]);
yeccpars2(765, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(766, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  addReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(767, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 876, [767 | __Ss], [__T | __Stack]);
yeccpars2(767, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(768, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 856, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(769, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  modReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(770, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  moveReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(771, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 850, [771 | __Ss], [__T | __Stack]);
yeccpars2(771, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(772, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 825, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(773, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  subtractReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(774, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 824, [774 | __Ss], [__T | __Stack]);
yeccpars2(774, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(775, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReplys, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(776, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 788, [776 | __Ss], [__T | __Stack]);
yeccpars2(776, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(777, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReplys, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(778, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 783, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(784, __Cat, [778 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(779, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {context,__1},
 yeccpars2(yeccgoto(commandReplys, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(780, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ActionReply'{errorDescriptor = __1},
 yeccpars2(yeccgoto(actionReplyBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(781, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReplys, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(782, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReplys, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(783, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'IEPSToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 772, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 771, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 767, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 768, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 773, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 769, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 770, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 766, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'ContextAttrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'EmergencyOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(784, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_action_reply([__1|__2]),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(actionReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(785, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 783, [785 | __Ss], [__T | __Stack]);
yeccpars2(785, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(787, __Cat, [785 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(786, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [{error,__2}],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(commandReplyList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(787, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(commandReplyList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(788, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(789, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 790, [789 | __Ss], [__T | __Stack]);
yeccpars2(789, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(791, __Cat, [789 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(790, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 797, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 796, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 793, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 795, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 794, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 792, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 486, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(791, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,#'AmmsReply'{terminationID = [__3], terminationAudit = __4}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(ammsReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(792, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [792 | __Ss], [__T | __Stack]);
yeccpars2(792, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mediaToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(793, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [793 | __Ss], [__T | __Stack]);
yeccpars2(793, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [793 | __Ss], [__T | __Stack]);
yeccpars2(793, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  modemToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(794, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [794 | __Ss], [__T | __Stack]);
yeccpars2(794, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  muxToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(795, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 682, [795 | __Ss], [__T | __Stack]);
yeccpars2(795, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  observedEventsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(796, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 817, [796 | __Ss], [__T | __Stack]);
yeccpars2(796, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  packagesToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(797, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [797 | __Ss], [__T | __Stack]);
yeccpars2(797, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  statsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(798, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditReturnItem,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(799, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 813, [799 | __Ss], [__T | __Stack]);
yeccpars2(799, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(814, __Cat, [799 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(800, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {digitMapDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(801, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {errorDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(802, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventBufferDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(803, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(804, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mediaDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(805, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(806, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {muxDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(807, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {observedEventsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(808, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {packagesDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(809, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signalsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(810, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {statisticsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(811, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 812, [811 | __Ss], [__T | __Stack]);
yeccpars2(811, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(812, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(ammsReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(813, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 797, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 796, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 793, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 795, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 794, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 792, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 486, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(814, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_terminationAudit([__1|__2]),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(terminationAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(815, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 813, [815 | __Ss], [__T | __Stack]);
yeccpars2(815, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(816, __Cat, [815 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(816, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditReturnParameterList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(817, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [817 | __Ss], [__T | __Stack]);
yeccpars2(817, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(818, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 819, [818 | __Ss], [__T | __Stack]);
yeccpars2(818, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(820, __Cat, [818 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(819, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(820, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 821, [820 | __Ss], [__T | __Stack]);
yeccpars2(820, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(821, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(packagesDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(822, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 819, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(823, __Cat, [822 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(823, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(packagesItems, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(824, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'ActionReply'.contextId,__5,__3),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(actionReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(825, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(826, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 827, [826 | __Ss], [__T | __Stack]);
yeccpars2(826, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  {serviceChangeResParms,#'ServiceChangeResParm'{}},
 yeccpars2(828, __Cat, [826 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(827, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 829, [827 | __Ss], [__T | __Stack]);
yeccpars2(827, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [827 | __Ss], [__T | __Stack]);
yeccpars2(827, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(828, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {serviceChangeReply,#'ServiceChangeReply'{terminationID = [__3], serviceChangeResult = __4}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(serviceChangeReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(829, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 834, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(830, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 833, [830 | __Ss], [__T | __Stack]);
yeccpars2(830, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(831, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 832, [831 | __Ss], [__T | __Stack]);
yeccpars2(831, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(832, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {serviceChangeResParms,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(833, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {errorDescriptor,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(834, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 838, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 836, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 835, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 837, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(835, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 666, [835 | __Ss], [__T | __Stack]);
yeccpars2(835, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(836, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 664, [836 | __Ss], [__T | __Stack]);
yeccpars2(836, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(837, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 658, [837 | __Ss], [__T | __Stack]);
yeccpars2(837, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(838, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 656, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(839, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 845, [839 | __Ss], [__T | __Stack]);
yeccpars2(839, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(846, __Cat, [839 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(840, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {address,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(841, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mgc_id,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(842, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {profile,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(843, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {version,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(844, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {time_stamp,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(845, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 459, [845 | __Ss], [__T | __Stack]);
yeccpars2(845, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 838, [845 | __Ss], [__T | __Stack]);
yeccpars2(845, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 836, [845 | __Ss], [__T | __Stack]);
yeccpars2(845, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 835, [845 | __Ss], [__T | __Stack]);
yeccpars2(845, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 837, [845 | __Ss], [__T | __Stack]);
yeccpars2(845, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(846, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 847, [846 | __Ss], [__T | __Stack]);
yeccpars2(846, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(847, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_ServiceChangeResParm([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(serviceChangeReplyDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(848, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 845, [848 | __Ss], [__T | __Stack]);
yeccpars2(848, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(849, __Cat, [848 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(849, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(servChgReplyParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(850, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(851, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 852, [851 | __Ss], [__T | __Stack]);
yeccpars2(851, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(853, __Cat, [851 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(852, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [852 | __Ss], [__T | __Stack]);
yeccpars2(852, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(853, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {notifyReply,#'NotifyReply'{terminationID = [__3], errorDescriptor = __4}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(notifyReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(854, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 855, [854 | __Ss], [__T | __Stack]);
yeccpars2(854, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(855, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(notifyReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(856, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 857, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(857, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 863, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(858, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditValueReply,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(859, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 860, [859 | __Ss], [__T | __Stack]);
yeccpars2(859, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditResult,#'AuditResult'{terminationID = __1, terminationAuditResult = []}},
 yeccpars2(yeccgoto(auditOther, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(860, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 797, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 796, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 793, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 795, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 794, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 792, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 486, [860 | __Ss], [__T | __Stack]);
yeccpars2(860, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(861, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 862, [861 | __Ss], [__T | __Stack]);
yeccpars2(861, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(862, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditResult,#'AuditResult'{terminationID = __1, terminationAuditResult = __3}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditOther, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(863, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 866, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [863 | __Ss], [__T | __Stack]);
yeccpars2(863, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(864, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditValueReply,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(865, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {contextAuditResult,__1},
 yeccpars2(yeccgoto(contextTerminationAudit, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(866, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 869, [866 | __Ss], [__T | __Stack]);
yeccpars2(866, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(867, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 868, [867 | __Ss], [__T | __Stack]);
yeccpars2(867, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(868, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {error,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(contextTerminationAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(869, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [869 | __Ss], [__T | __Stack]);
yeccpars2(869, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(870, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 872, [870 | __Ss], [__T | __Stack]);
yeccpars2(870, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(871, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint(__1,0,999),
 yeccpars2(yeccgoto(errorCode, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(872, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 873, [872 | __Ss], [__T | __Stack]);
yeccpars2(872, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(874, __Cat, [872 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(873, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  value_of(__1),
 yeccpars2(yeccgoto(errorText, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(874, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 875, [874 | __Ss], [__T | __Stack]);
yeccpars2(874, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(875, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ErrorDescriptor'{errorCode = __3, errorText = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(errorDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(876, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 877, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [876 | __Ss], [__T | __Stack]);
yeccpars2(876, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(877, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 863, [877 | __Ss], [__T | __Stack]);
yeccpars2(877, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(878, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditCapReply,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(879, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditCapReply,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(880, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'NULL',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(optImmAckRequired, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(881, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [881 | __Ss], [__T | __Stack]);
yeccpars2(881, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(882, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 883, [882 | __Ss], [__T | __Stack]);
yeccpars2(882, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(883, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 884, [883 | __Ss], [__T | __Stack]);
yeccpars2(883, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(884, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionPending'{transactionId = ensure_transactionID(__3)},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(transactionPending, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(885, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 885, [885 | __Ss], [__T | __Stack]);
yeccpars2(885, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(893, __Cat, [885 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(886, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 889, [886 | __Ss], [__T | __Stack]);
yeccpars2(886, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(887, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 885, [887 | __Ss], [__T | __Stack]);
yeccpars2(887, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(888, __Cat, [887 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(888, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(daddr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(889, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 890, [889 | __Ss], [__T | __Stack]);
yeccpars2(889, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainAddress(__2,asn1_NOVALUE),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(domainAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(890, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [890 | __Ss], [__T | __Stack]);
yeccpars2(890, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(891, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [891 | __Ss], [__T | __Stack]);
yeccpars2(891, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(892, __Cat, [891 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(892, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainAddress(__2,__5),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(domainAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(893, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [colon|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(daddr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(894, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 895, [894 | __Ss], [__T | __Stack]);
yeccpars2(894, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(895, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 896, [895 | __Ss], [__T | __Stack]);
yeccpars2(895, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainName(__2,asn1_NOVALUE),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(domainName, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(896, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [896 | __Ss], [__T | __Stack]);
yeccpars2(896, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(897, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [897 | __Ss], [__T | __Stack]);
yeccpars2(897, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(898, __Cat, [897 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(898, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainName(__2,__5),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(domainName, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(899, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'MegacoMessage'{authHeader = __2, mess = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(megacoMessage, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(900, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [900 | __Ss], [__T | __Stack]);
yeccpars2(900, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(901, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 902, [901 | __Ss], [__T | __Stack]);
yeccpars2(901, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(902, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [902 | __Ss], [__T | __Stack]);
yeccpars2(902, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(903, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 904, [903 | __Ss], [__T | __Stack]);
yeccpars2(903, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(904, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'RequestIDToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'DirectionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [904 | __Ss], [__T | __Stack]);
yeccpars2(904, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(905, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [905 | __Ss], [__T | __Stack]);
yeccpars2(905, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(906, __Cat, [905 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(906, __Cat, __Ss,  [__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_auth_header(__3,__5,__7),
 __Nss = lists:nthtail(7, __Ss),
 yeccpars2(yeccgoto(authenticationHeader, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto(actionReply, 753) ->
 755;
yeccgoto(actionReply, 759) ->
 761;
yeccgoto(actionReplyBody, 765) ->
 774;
yeccgoto(actionReplyList, 755) ->
 760;
yeccgoto(actionReplyList, 761) ->
 762;
yeccgoto(actionRequest, 116) ->
 118;
yeccgoto(actionRequest, 119) ->
 122;
yeccgoto(actionRequest, 731) ->
 738;
yeccgoto(actionRequest, 734) ->
 735;
yeccgoto(actionRequestBody, 127) ->
 143;
yeccgoto(actionRequestItem, 127) ->
 144;
yeccgoto(actionRequestItem, 594) ->
 596;
yeccgoto(actionRequestItems, 144) ->
 595;
yeccgoto(actionRequestItems, 596) ->
 597;
yeccgoto(actionRequestList, 118) ->
 120;
yeccgoto(actionRequestList, 122) ->
 123;
yeccgoto(actionRequestList, 735) ->
 736;
yeccgoto(actionRequestList, 738) ->
 739;
yeccgoto(alternativeValue, 227) ->
 237;
yeccgoto(ammParameter, 161) ->
 172;
yeccgoto(ammParameter, 182) ->
 185;
yeccgoto(ammParameters, 172) ->
 183;
yeccgoto(ammParameters, 185) ->
 186;
yeccgoto(ammRequest, 127) ->
 145;
yeccgoto(ammRequest, 594) ->
 145;
yeccgoto(ammRequestBody, 160) ->
 162;
yeccgoto(ammToken, 127) ->
 146;
yeccgoto(ammToken, 594) ->
 146;
yeccgoto(ammsReply, 765) ->
 775;
yeccgoto(ammsReply, 783) ->
 775;
yeccgoto(ammsReplyBody, 789) ->
 791;
yeccgoto(ammsToken, 765) ->
 776;
yeccgoto(ammsToken, 783) ->
 776;
yeccgoto(auditDescriptor, 161) ->
 173;
yeccgoto(auditDescriptor, 182) ->
 173;
yeccgoto(auditDescriptor, 618) ->
 620;
yeccgoto(auditDescriptorBody, 484) ->
 496;
yeccgoto(auditItem, 484) ->
 497;
yeccgoto(auditItem, 589) ->
 591;
yeccgoto(auditItem, 628) ->
 637;
yeccgoto(auditItem, 650) ->
 637;
yeccgoto(auditItemList, 497) ->
 590;
yeccgoto(auditItemList, 591) ->
 592;
yeccgoto(auditOther, 856) ->
 858;
yeccgoto(auditOther, 876) ->
 878;
yeccgoto(auditReply, 765) ->
 777;
yeccgoto(auditReply, 783) ->
 777;
yeccgoto(auditRequest, 127) ->
 147;
yeccgoto(auditRequest, 594) ->
 147;
yeccgoto(auditReturnItem, 484) ->
 498;
yeccgoto(auditReturnItem, 589) ->
 498;
yeccgoto(auditReturnItem, 628) ->
 498;
yeccgoto(auditReturnItem, 650) ->
 498;
yeccgoto(auditReturnItem, 790) ->
 798;
yeccgoto(auditReturnItem, 813) ->
 798;
yeccgoto(auditReturnItem, 860) ->
 798;
yeccgoto(auditReturnParameter, 790) ->
 799;
yeccgoto(auditReturnParameter, 813) ->
 815;
yeccgoto(auditReturnParameter, 860) ->
 799;
yeccgoto(auditReturnParameterList, 799) ->
 814;
yeccgoto(auditReturnParameterList, 815) ->
 816;
yeccgoto(authenticationHeader, 3) ->
 5;
yeccgoto(commandReplyList, 778) ->
 784;
yeccgoto(commandReplyList, 785) ->
 787;
yeccgoto(commandReplys, 765) ->
 778;
yeccgoto(commandReplys, 783) ->
 785;
yeccgoto(commandRequest, 127) ->
 148;
yeccgoto(commandRequest, 594) ->
 148;
yeccgoto(contextAttrDescriptor, 127) ->
 149;
yeccgoto(contextAttrDescriptor, 594) ->
 149;
yeccgoto(contextAttrDescriptor, 765) ->
 149;
yeccgoto(contextAttrDescriptor, 783) ->
 149;
yeccgoto(contextAudit, 127) ->
 150;
yeccgoto(contextAudit, 594) ->
 150;
yeccgoto(contextAuditProperties, 702) ->
 705;
yeccgoto(contextAuditProperties, 707) ->
 708;
yeccgoto(contextAuditProperty, 697) ->
 702;
yeccgoto(contextAuditProperty, 704) ->
 707;
yeccgoto(contextID, 124) ->
 125;
yeccgoto(contextID, 718) ->
 719;
yeccgoto(contextID, 720) ->
 723;
yeccgoto(contextID, 763) ->
 764;
yeccgoto(contextIDs, 719) ->
 721;
yeccgoto(contextIDs, 723) ->
 724;
yeccgoto(contextIdList, 709) ->
 711;
yeccgoto(contextProperty, 127) ->
 151;
yeccgoto(contextProperty, 594) ->
 151;
yeccgoto(contextProperty, 765) ->
 779;
yeccgoto(contextProperty, 783) ->
 779;
yeccgoto(contextTerminationAudit, 857) ->
 864;
yeccgoto(contextTerminationAudit, 877) ->
 879;
yeccgoto(daddr, 89) ->
 886;
yeccgoto(daddr, 885) ->
 893;
yeccgoto(daddr, 887) ->
 888;
yeccgoto(deviceName, 93) ->
 95;
yeccgoto(digitMapDescriptor, 161) ->
 174;
yeccgoto(digitMapDescriptor, 182) ->
 174;
yeccgoto(digitMapDescriptor, 790) ->
 800;
yeccgoto(digitMapDescriptor, 813) ->
 800;
yeccgoto(digitMapDescriptor, 860) ->
 800;
yeccgoto(direction, 276) ->
 280;
yeccgoto(domainAddress, 87) ->
 90;
yeccgoto(domainAddress, 658) ->
 90;
yeccgoto(domainAddress, 666) ->
 90;
yeccgoto(domainName, 87) ->
 91;
yeccgoto(domainName, 658) ->
 91;
yeccgoto(domainName, 666) ->
 91;
yeccgoto(embedFirst, 423) ->
 425;
yeccgoto(embedFirst, 427) ->
 429;
yeccgoto(embedNoSig, 405) ->
 410;
yeccgoto(embedNoSig, 418) ->
 410;
yeccgoto(embedSig, 442) ->
 446;
yeccgoto(embedSig, 450) ->
 446;
yeccgoto(embedWithSig, 405) ->
 411;
yeccgoto(embedWithSig, 418) ->
 411;
yeccgoto(errorCode, 869) ->
 870;
yeccgoto(errorDescriptor, 92) ->
 106;
yeccgoto(errorDescriptor, 676) ->
 678;
yeccgoto(errorDescriptor, 753) ->
 756;
yeccgoto(errorDescriptor, 765) ->
 780;
yeccgoto(errorDescriptor, 783) ->
 786;
yeccgoto(errorDescriptor, 790) ->
 801;
yeccgoto(errorDescriptor, 813) ->
 801;
yeccgoto(errorDescriptor, 827) ->
 830;
yeccgoto(errorDescriptor, 852) ->
 854;
yeccgoto(errorDescriptor, 860) ->
 801;
yeccgoto(errorDescriptor, 863) ->
 867;
yeccgoto(errorText, 872) ->
 874;
yeccgoto(eventBufferControl, 341) ->
 344;
yeccgoto(eventBufferControl, 348) ->
 344;
yeccgoto(eventBufferControlState, 358) ->
 361;
yeccgoto(eventBufferDescriptor, 161) ->
 175;
yeccgoto(eventBufferDescriptor, 182) ->
 175;
yeccgoto(eventBufferDescriptor, 790) ->
 802;
yeccgoto(eventBufferDescriptor, 813) ->
 802;
yeccgoto(eventBufferDescriptor, 860) ->
 802;
yeccgoto(eventDM, 405) ->
 412;
yeccgoto(eventDM, 418) ->
 412;
yeccgoto(eventDM, 442) ->
 447;
yeccgoto(eventDM, 450) ->
 447;
yeccgoto(eventParameter, 405) ->
 413;
yeccgoto(eventParameter, 418) ->
 421;
yeccgoto(eventParameterName, 405) ->
 414;
yeccgoto(eventParameterName, 418) ->
 414;
yeccgoto(eventParameterName, 442) ->
 414;
yeccgoto(eventParameterName, 450) ->
 414;
yeccgoto(eventParameterName, 468) ->
 414;
yeccgoto(eventParameterName, 472) ->
 414;
yeccgoto(eventParameterName, 579) ->
 582;
yeccgoto(eventParameters, 413) ->
 419;
yeccgoto(eventParameters, 421) ->
 422;
yeccgoto(eventSpec, 458) ->
 460;
yeccgoto(eventSpec, 479) ->
 482;
yeccgoto(eventSpecList, 460) ->
 480;
yeccgoto(eventSpecList, 482) ->
 483;
yeccgoto(eventStream, 579) ->
 583;
yeccgoto(eventStreamOrOther, 405) ->
 415;
yeccgoto(eventStreamOrOther, 418) ->
 415;
yeccgoto(eventStreamOrOther, 442) ->
 448;
yeccgoto(eventStreamOrOther, 450) ->
 448;
yeccgoto(eventStreamOrOther, 468) ->
 470;
yeccgoto(eventStreamOrOther, 472) ->
 470;
yeccgoto(eventsDescriptor, 161) ->
 176;
yeccgoto(eventsDescriptor, 182) ->
 176;
yeccgoto(eventsDescriptor, 790) ->
 803;
yeccgoto(eventsDescriptor, 813) ->
 803;
yeccgoto(eventsDescriptor, 860) ->
 803;
yeccgoto(extension, 628) ->
 638;
yeccgoto(extension, 650) ->
 638;
yeccgoto(extensionParameter, 628) ->
 639;
yeccgoto(extensionParameter, 650) ->
 639;
yeccgoto(iepsValue, 127) ->
 152;
yeccgoto(iepsValue, 594) ->
 152;
yeccgoto(iepsValue, 765) ->
 152;
yeccgoto(iepsValue, 783) ->
 152;
yeccgoto(indAudauditReturnParameter, 484) ->
 499;
yeccgoto(indAudauditReturnParameter, 508) ->
 516;
yeccgoto(indAudauditReturnParameter, 589) ->
 499;
yeccgoto(indAudauditReturnParameter, 628) ->
 499;
yeccgoto(indAudauditReturnParameter, 650) ->
 499;
yeccgoto(indAudcontextAttrDescriptor, 693) ->
 695;
yeccgoto(indAuddigitMapDescriptor, 484) ->
 500;
yeccgoto(indAuddigitMapDescriptor, 508) ->
 500;
yeccgoto(indAuddigitMapDescriptor, 589) ->
 500;
yeccgoto(indAuddigitMapDescriptor, 628) ->
 500;
yeccgoto(indAuddigitMapDescriptor, 650) ->
 500;
yeccgoto(indAudeventBufferDescriptor, 484) ->
 501;
yeccgoto(indAudeventBufferDescriptor, 508) ->
 501;
yeccgoto(indAudeventBufferDescriptor, 589) ->
 501;
yeccgoto(indAudeventBufferDescriptor, 628) ->
 501;
yeccgoto(indAudeventBufferDescriptor, 650) ->
 501;
yeccgoto(indAudeventSpec, 576) ->
 577;
yeccgoto(indAudeventSpecParameter, 579) ->
 584;
yeccgoto(indAudeventsDescriptor, 484) ->
 502;
yeccgoto(indAudeventsDescriptor, 508) ->
 502;
yeccgoto(indAudeventsDescriptor, 589) ->
 502;
yeccgoto(indAudeventsDescriptor, 628) ->
 502;
yeccgoto(indAudeventsDescriptor, 650) ->
 502;
yeccgoto(indAudlocalControlDescriptor, 538) ->
 542;
yeccgoto(indAudlocalControlDescriptor, 548) ->
 542;
yeccgoto(indAudlocalControlDescriptor, 559) ->
 542;
yeccgoto(indAudlocalParm, 562) ->
 563;
yeccgoto(indAudlocalParm, 565) ->
 568;
yeccgoto(indAudlocalParmList, 563) ->
 566;
yeccgoto(indAudlocalParmList, 568) ->
 569;
yeccgoto(indAudmediaDescriptor, 484) ->
 503;
yeccgoto(indAudmediaDescriptor, 508) ->
 503;
yeccgoto(indAudmediaDescriptor, 589) ->
 503;
yeccgoto(indAudmediaDescriptor, 628) ->
 503;
yeccgoto(indAudmediaDescriptor, 650) ->
 503;
yeccgoto(indAudmediaParm, 538) ->
 543;
yeccgoto(indAudmediaParm, 548) ->
 551;
yeccgoto(indAudmediaParms, 543) ->
 549;
yeccgoto(indAudmediaParms, 551) ->
 552;
yeccgoto(indAudpackagesDescriptor, 484) ->
 504;
yeccgoto(indAudpackagesDescriptor, 508) ->
 504;
yeccgoto(indAudpackagesDescriptor, 589) ->
 504;
yeccgoto(indAudpackagesDescriptor, 628) ->
 504;
yeccgoto(indAudpackagesDescriptor, 650) ->
 504;
yeccgoto(indAudrequestedEvent, 572) ->
 573;
yeccgoto(indAudsignalList, 521) ->
 525;
yeccgoto(indAudsignalParm, 521) ->
 526;
yeccgoto(indAudsignalsDescriptor, 484) ->
 505;
yeccgoto(indAudsignalsDescriptor, 508) ->
 505;
yeccgoto(indAudsignalsDescriptor, 589) ->
 505;
yeccgoto(indAudsignalsDescriptor, 628) ->
 505;
yeccgoto(indAudsignalsDescriptor, 650) ->
 505;
yeccgoto(indAudstatisticsDescriptor, 484) ->
 506;
yeccgoto(indAudstatisticsDescriptor, 508) ->
 506;
yeccgoto(indAudstatisticsDescriptor, 538) ->
 544;
yeccgoto(indAudstatisticsDescriptor, 548) ->
 544;
yeccgoto(indAudstatisticsDescriptor, 559) ->
 544;
yeccgoto(indAudstatisticsDescriptor, 589) ->
 506;
yeccgoto(indAudstatisticsDescriptor, 628) ->
 506;
yeccgoto(indAudstatisticsDescriptor, 650) ->
 506;
yeccgoto(indAudstreamDescriptor, 538) ->
 545;
yeccgoto(indAudstreamDescriptor, 548) ->
 545;
yeccgoto(indAudstreamParm, 538) ->
 546;
yeccgoto(indAudstreamParm, 548) ->
 546;
yeccgoto(indAudstreamParm, 559) ->
 560;
yeccgoto(indAudterminationAudit, 484) ->
 507;
yeccgoto(indAudterminationAudit, 589) ->
 507;
yeccgoto(indAudterminationAudit, 628) ->
 507;
yeccgoto(indAudterminationAudit, 650) ->
 507;
yeccgoto(indAudterminationAuditList, 499) ->
 509;
yeccgoto(indAudterminationAuditList, 516) ->
 517;
yeccgoto(indAudterminationStateDescriptor, 538) ->
 547;
yeccgoto(indAudterminationStateDescriptor, 548) ->
 547;
yeccgoto(indAudterminationStateParm, 553) ->
 554;
yeccgoto(localControlDescriptor, 324) ->
 330;
yeccgoto(localControlDescriptor, 336) ->
 330;
yeccgoto(localControlDescriptor, 364) ->
 330;
yeccgoto(localControlDescriptor, 366) ->
 330;
yeccgoto(localParm, 371) ->
 375;
yeccgoto(localParm, 377) ->
 380;
yeccgoto(localParmList, 375) ->
 378;
yeccgoto(localParmList, 380) ->
 381;
yeccgoto(mId, 87) ->
 92;
yeccgoto(mId, 658) ->
 659;
yeccgoto(mId, 666) ->
 667;
yeccgoto(mediaDescriptor, 161) ->
 177;
yeccgoto(mediaDescriptor, 182) ->
 177;
yeccgoto(mediaDescriptor, 790) ->
 804;
yeccgoto(mediaDescriptor, 813) ->
 804;
yeccgoto(mediaDescriptor, 860) ->
 804;
yeccgoto(mediaParm, 324) ->
 331;
yeccgoto(mediaParm, 336) ->
 339;
yeccgoto(mediaParmList, 331) ->
 337;
yeccgoto(mediaParmList, 339) ->
 340;
yeccgoto(megacoMessage, 0) ->
 2;
yeccgoto(message, 5) ->
 86;
yeccgoto(messageBody, 92) ->
 107;
yeccgoto(modemDescriptor, 161) ->
 178;
yeccgoto(modemDescriptor, 182) ->
 178;
yeccgoto(modemDescriptor, 790) ->
 805;
yeccgoto(modemDescriptor, 813) ->
 805;
yeccgoto(modemDescriptor, 860) ->
 805;
yeccgoto(modemType, 303) ->
 322;
yeccgoto(modemType, 304) ->
 305;
yeccgoto(modemType, 307) ->
 320;
yeccgoto(modemTypeList, 305) ->
 308;
yeccgoto(modemTypeList, 320) ->
 321;
yeccgoto(mtpAddress, 93) ->
 96;
yeccgoto(muxDescriptor, 161) ->
 179;
yeccgoto(muxDescriptor, 182) ->
 179;
yeccgoto(muxDescriptor, 790) ->
 806;
yeccgoto(muxDescriptor, 813) ->
 806;
yeccgoto(muxDescriptor, 860) ->
 806;
yeccgoto(muxType, 292) ->
 293;
yeccgoto(notificationReason, 263) ->
 268;
yeccgoto(notificationReason, 269) ->
 272;
yeccgoto(notificationReasons, 268) ->
 270;
yeccgoto(notificationReasons, 272) ->
 273;
yeccgoto(notifyReply, 765) ->
 781;
yeccgoto(notifyReply, 783) ->
 781;
yeccgoto(notifyReplyBody, 851) ->
 853;
yeccgoto(notifyRequest, 127) ->
 153;
yeccgoto(notifyRequest, 594) ->
 153;
yeccgoto(notifyRequestBody, 676) ->
 679;
yeccgoto(observedEvent, 458) ->
 461;
yeccgoto(observedEvent, 479) ->
 461;
yeccgoto(observedEvent, 684) ->
 685;
yeccgoto(observedEvent, 686) ->
 689;
yeccgoto(observedEventBody, 467) ->
 469;
yeccgoto(observedEventBody, 477) ->
 478;
yeccgoto(observedEventParameter, 468) ->
 471;
yeccgoto(observedEventParameter, 472) ->
 475;
yeccgoto(observedEventParameters, 471) ->
 473;
yeccgoto(observedEventParameters, 475) ->
 476;
yeccgoto(observedEvents, 685) ->
 687;
yeccgoto(observedEvents, 689) ->
 690;
yeccgoto(observedEventsDescriptor, 676) ->
 680;
yeccgoto(observedEventsDescriptor, 790) ->
 807;
yeccgoto(observedEventsDescriptor, 813) ->
 807;
yeccgoto(observedEventsDescriptor, 860) ->
 807;
yeccgoto(onOrOff, 382) ->
 385;
yeccgoto(onOrOff, 386) ->
 387;
yeccgoto(onOrOff, 691) ->
 692;
yeccgoto(optAuditDescriptor, 617) ->
 619;
yeccgoto(optAuditDescriptor, 726) ->
 727;
yeccgoto(optAuditDescriptor, 729) ->
 730;
yeccgoto(optImmAckRequired, 751) ->
 753;
yeccgoto(optIndAudeventSpecParameter, 578) ->
 580;
yeccgoto(optIndAudsignalParm, 494) ->
 522;
yeccgoto(optIndAudsignalParm, 514) ->
 522;
yeccgoto(optPropertyParms, 309) ->
 311;
yeccgoto(optPropertyParms, 322) ->
 323;
yeccgoto(optSep, 0) ->
 3;
yeccgoto(optSep, 87) ->
 93;
yeccgoto(optSep, 95) ->
 100;
yeccgoto(optSep, 96) ->
 99;
yeccgoto(optSep, 458) ->
 462;
yeccgoto(optSep, 463) ->
 464;
yeccgoto(optSep, 465) ->
 466;
yeccgoto(optSep, 479) ->
 462;
yeccgoto(optSep, 658) ->
 93;
yeccgoto(optSep, 666) ->
 93;
yeccgoto(optSep, 684) ->
 462;
yeccgoto(optSep, 686) ->
 462;
yeccgoto(optSep, 891) ->
 892;
yeccgoto(optSep, 897) ->
 898;
yeccgoto(optSep, 905) ->
 906;
yeccgoto(packagesDescriptor, 790) ->
 808;
yeccgoto(packagesDescriptor, 813) ->
 808;
yeccgoto(packagesDescriptor, 860) ->
 808;
yeccgoto(packagesItem, 534) ->
 535;
yeccgoto(packagesItem, 817) ->
 818;
yeccgoto(packagesItem, 819) ->
 822;
yeccgoto(packagesItems, 818) ->
 820;
yeccgoto(packagesItems, 822) ->
 823;
yeccgoto(parmValue, 220) ->
 231;
yeccgoto(parmValue, 312) ->
 319;
yeccgoto(parmValue, 414) ->
 417;
yeccgoto(parmValue, 639) ->
 655;
yeccgoto(pathName, 93) ->
 97;
yeccgoto(pkgdName, 187) ->
 188;
yeccgoto(pkgdName, 191) ->
 188;
yeccgoto(pkgdName, 200) ->
 202;
yeccgoto(pkgdName, 207) ->
 202;
yeccgoto(pkgdName, 284) ->
 202;
yeccgoto(pkgdName, 287) ->
 202;
yeccgoto(pkgdName, 310) ->
 312;
yeccgoto(pkgdName, 314) ->
 312;
yeccgoto(pkgdName, 341) ->
 312;
yeccgoto(pkgdName, 348) ->
 312;
yeccgoto(pkgdName, 371) ->
 312;
yeccgoto(pkgdName, 377) ->
 312;
yeccgoto(pkgdName, 397) ->
 398;
yeccgoto(pkgdName, 400) ->
 398;
yeccgoto(pkgdName, 434) ->
 435;
yeccgoto(pkgdName, 437) ->
 435;
yeccgoto(pkgdName, 462) ->
 477;
yeccgoto(pkgdName, 466) ->
 467;
yeccgoto(pkgdName, 518) ->
 519;
yeccgoto(pkgdName, 521) ->
 202;
yeccgoto(pkgdName, 531) ->
 202;
yeccgoto(pkgdName, 572) ->
 574;
yeccgoto(pkgdName, 576) ->
 578;
yeccgoto(pkgdName, 697) ->
 703;
yeccgoto(pkgdName, 704) ->
 703;
yeccgoto(pkgdName, 709) ->
 312;
yeccgoto(portNumber, 658) ->
 660;
yeccgoto(portNumber, 890) ->
 891;
yeccgoto(portNumber, 896) ->
 897;
yeccgoto(priority, 127) ->
 154;
yeccgoto(priority, 594) ->
 154;
yeccgoto(priority, 765) ->
 154;
yeccgoto(priority, 783) ->
 154;
yeccgoto(propertyParm, 310) ->
 313;
yeccgoto(propertyParm, 314) ->
 317;
yeccgoto(propertyParm, 341) ->
 345;
yeccgoto(propertyParm, 348) ->
 345;
yeccgoto(propertyParm, 371) ->
 376;
yeccgoto(propertyParm, 377) ->
 376;
yeccgoto(propertyParm, 709) ->
 712;
yeccgoto(propertyParmList, 313) ->
 315;
yeccgoto(propertyParmList, 317) ->
 318;
yeccgoto(propertyParmList, 712) ->
 715;
yeccgoto(propertyParms, 709) ->
 713;
yeccgoto(requestID, 259) ->
 260;
yeccgoto(requestID, 395) ->
 396;
yeccgoto(requestID, 432) ->
 433;
yeccgoto(requestID, 570) ->
 571;
yeccgoto(requestID, 682) ->
 683;
yeccgoto(requestedEvent, 397) ->
 399;
yeccgoto(requestedEvent, 400) ->
 403;
yeccgoto(requestedEventBody, 398) ->
 406;
yeccgoto(requestedEvents, 399) ->
 401;
yeccgoto(requestedEvents, 403) ->
 404;
yeccgoto(safeToken, 5) ->
 87;
yeccgoto(safeToken, 88) ->
 894;
yeccgoto(safeToken, 89) ->
 887;
yeccgoto(safeToken, 93) ->
 98;
yeccgoto(safeToken, 115) ->
 732;
yeccgoto(safeToken, 124) ->
 126;
yeccgoto(safeToken, 158) ->
 159;
yeccgoto(safeToken, 187) ->
 189;
yeccgoto(safeToken, 191) ->
 189;
yeccgoto(safeToken, 196) ->
 198;
yeccgoto(safeToken, 200) ->
 189;
yeccgoto(safeToken, 207) ->
 189;
yeccgoto(safeToken, 212) ->
 220;
yeccgoto(safeToken, 222) ->
 220;
yeccgoto(safeToken, 227) ->
 198;
yeccgoto(safeToken, 228) ->
 198;
yeccgoto(safeToken, 229) ->
 198;
yeccgoto(safeToken, 230) ->
 198;
yeccgoto(safeToken, 235) ->
 198;
yeccgoto(safeToken, 236) ->
 198;
yeccgoto(safeToken, 240) ->
 198;
yeccgoto(safeToken, 241) ->
 198;
yeccgoto(safeToken, 251) ->
 252;
yeccgoto(safeToken, 259) ->
 261;
yeccgoto(safeToken, 274) ->
 275;
yeccgoto(safeToken, 281) ->
 282;
yeccgoto(safeToken, 284) ->
 189;
yeccgoto(safeToken, 287) ->
 189;
yeccgoto(safeToken, 292) ->
 294;
yeccgoto(safeToken, 295) ->
 159;
yeccgoto(safeToken, 298) ->
 159;
yeccgoto(safeToken, 303) ->
 306;
yeccgoto(safeToken, 304) ->
 306;
yeccgoto(safeToken, 307) ->
 306;
yeccgoto(safeToken, 310) ->
 189;
yeccgoto(safeToken, 314) ->
 189;
yeccgoto(safeToken, 341) ->
 189;
yeccgoto(safeToken, 348) ->
 189;
yeccgoto(safeToken, 362) ->
 252;
yeccgoto(safeToken, 371) ->
 189;
yeccgoto(safeToken, 377) ->
 189;
yeccgoto(safeToken, 395) ->
 261;
yeccgoto(safeToken, 397) ->
 189;
yeccgoto(safeToken, 400) ->
 189;
yeccgoto(safeToken, 405) ->
 416;
yeccgoto(safeToken, 418) ->
 416;
yeccgoto(safeToken, 432) ->
 261;
yeccgoto(safeToken, 434) ->
 189;
yeccgoto(safeToken, 437) ->
 189;
yeccgoto(safeToken, 442) ->
 416;
yeccgoto(safeToken, 450) ->
 416;
yeccgoto(safeToken, 462) ->
 189;
yeccgoto(safeToken, 466) ->
 189;
yeccgoto(safeToken, 468) ->
 416;
yeccgoto(safeToken, 472) ->
 416;
yeccgoto(safeToken, 518) ->
 189;
yeccgoto(safeToken, 521) ->
 189;
yeccgoto(safeToken, 529) ->
 282;
yeccgoto(safeToken, 531) ->
 189;
yeccgoto(safeToken, 534) ->
 536;
yeccgoto(safeToken, 553) ->
 555;
yeccgoto(safeToken, 557) ->
 252;
yeccgoto(safeToken, 562) ->
 564;
yeccgoto(safeToken, 565) ->
 564;
yeccgoto(safeToken, 570) ->
 261;
yeccgoto(safeToken, 572) ->
 189;
yeccgoto(safeToken, 576) ->
 189;
yeccgoto(safeToken, 579) ->
 416;
yeccgoto(safeToken, 586) ->
 252;
yeccgoto(safeToken, 599) ->
 159;
yeccgoto(safeToken, 603) ->
 159;
yeccgoto(safeToken, 608) ->
 159;
yeccgoto(safeToken, 616) ->
 159;
yeccgoto(safeToken, 622) ->
 159;
yeccgoto(safeToken, 628) ->
 640;
yeccgoto(safeToken, 650) ->
 640;
yeccgoto(safeToken, 656) ->
 657;
yeccgoto(safeToken, 658) ->
 661;
yeccgoto(safeToken, 662) ->
 198;
yeccgoto(safeToken, 664) ->
 665;
yeccgoto(safeToken, 668) ->
 669;
yeccgoto(safeToken, 670) ->
 671;
yeccgoto(safeToken, 672) ->
 673;
yeccgoto(safeToken, 674) ->
 159;
yeccgoto(safeToken, 682) ->
 261;
yeccgoto(safeToken, 697) ->
 189;
yeccgoto(safeToken, 704) ->
 189;
yeccgoto(safeToken, 709) ->
 189;
yeccgoto(safeToken, 718) ->
 126;
yeccgoto(safeToken, 720) ->
 126;
yeccgoto(safeToken, 725) ->
 159;
yeccgoto(safeToken, 728) ->
 159;
yeccgoto(safeToken, 741) ->
 742;
yeccgoto(safeToken, 744) ->
 742;
yeccgoto(safeToken, 749) ->
 732;
yeccgoto(safeToken, 763) ->
 126;
yeccgoto(safeToken, 788) ->
 159;
yeccgoto(safeToken, 817) ->
 536;
yeccgoto(safeToken, 819) ->
 536;
yeccgoto(safeToken, 825) ->
 159;
yeccgoto(safeToken, 850) ->
 159;
yeccgoto(safeToken, 856) ->
 159;
yeccgoto(safeToken, 863) ->
 159;
yeccgoto(safeToken, 869) ->
 871;
yeccgoto(safeToken, 876) ->
 159;
yeccgoto(safeToken, 881) ->
 732;
yeccgoto(safeToken, 885) ->
 887;
yeccgoto(safeToken, 887) ->
 887;
yeccgoto(safeToken, 890) ->
 661;
yeccgoto(safeToken, 896) ->
 661;
yeccgoto(safeToken, 900) ->
 901;
yeccgoto(safeToken, 902) ->
 903;
yeccgoto(safeToken, 904) ->
 905;
yeccgoto(secondEventParameter, 442) ->
 449;
yeccgoto(secondEventParameter, 450) ->
 453;
yeccgoto(secondEventParameters, 449) ->
 451;
yeccgoto(secondEventParameters, 453) ->
 454;
yeccgoto(secondRequestedEvent, 434) ->
 436;
yeccgoto(secondRequestedEvent, 437) ->
 440;
yeccgoto(secondRequestedEventBody, 435) ->
 443;
yeccgoto(secondRequestedEvents, 436) ->
 438;
yeccgoto(secondRequestedEvents, 440) ->
 441;
yeccgoto(servChgReplyParm, 834) ->
 839;
yeccgoto(servChgReplyParm, 845) ->
 848;
yeccgoto(servChgReplyParms, 839) ->
 846;
yeccgoto(servChgReplyParms, 848) ->
 849;
yeccgoto(serviceChangeAddress, 628) ->
 641;
yeccgoto(serviceChangeAddress, 650) ->
 641;
yeccgoto(serviceChangeAddress, 834) ->
 840;
yeccgoto(serviceChangeAddress, 845) ->
 840;
yeccgoto(serviceChangeDelay, 628) ->
 642;
yeccgoto(serviceChangeDelay, 650) ->
 642;
yeccgoto(serviceChangeDescriptor, 624) ->
 626;
yeccgoto(serviceChangeMethod, 628) ->
 643;
yeccgoto(serviceChangeMethod, 650) ->
 643;
yeccgoto(serviceChangeMgcId, 628) ->
 644;
yeccgoto(serviceChangeMgcId, 650) ->
 644;
yeccgoto(serviceChangeMgcId, 834) ->
 841;
yeccgoto(serviceChangeMgcId, 845) ->
 841;
yeccgoto(serviceChangeParm, 628) ->
 645;
yeccgoto(serviceChangeParm, 650) ->
 653;
yeccgoto(serviceChangeParms, 645) ->
 651;
yeccgoto(serviceChangeParms, 653) ->
 654;
yeccgoto(serviceChangeProfile, 628) ->
 646;
yeccgoto(serviceChangeProfile, 650) ->
 646;
yeccgoto(serviceChangeProfile, 834) ->
 842;
yeccgoto(serviceChangeProfile, 845) ->
 842;
yeccgoto(serviceChangeReason, 628) ->
 647;
yeccgoto(serviceChangeReason, 650) ->
 647;
yeccgoto(serviceChangeReply, 765) ->
 782;
yeccgoto(serviceChangeReply, 783) ->
 782;
yeccgoto(serviceChangeReplyBody, 826) ->
 828;
yeccgoto(serviceChangeReplyDescriptor, 827) ->
 831;
yeccgoto(serviceChangeRequest, 127) ->
 155;
yeccgoto(serviceChangeRequest, 594) ->
 155;
yeccgoto(serviceChangeVersion, 628) ->
 648;
yeccgoto(serviceChangeVersion, 650) ->
 648;
yeccgoto(serviceChangeVersion, 834) ->
 843;
yeccgoto(serviceChangeVersion, 845) ->
 843;
yeccgoto(serviceState, 353) ->
 357;
yeccgoto(serviceStates, 341) ->
 346;
yeccgoto(serviceStates, 348) ->
 346;
yeccgoto(sigParameter, 212) ->
 221;
yeccgoto(sigParameter, 222) ->
 225;
yeccgoto(sigParameters, 221) ->
 223;
yeccgoto(sigParameters, 225) ->
 226;
yeccgoto(signalList, 200) ->
 203;
yeccgoto(signalList, 207) ->
 203;
yeccgoto(signalListId, 281) ->
 283;
yeccgoto(signalListId, 529) ->
 530;
yeccgoto(signalListParm, 284) ->
 285;
yeccgoto(signalListParm, 287) ->
 290;
yeccgoto(signalListParm, 531) ->
 532;
yeccgoto(signalListParms, 285) ->
 288;
yeccgoto(signalListParms, 290) ->
 291;
yeccgoto(signalName, 200) ->
 204;
yeccgoto(signalName, 207) ->
 204;
yeccgoto(signalName, 284) ->
 204;
yeccgoto(signalName, 287) ->
 204;
yeccgoto(signalName, 521) ->
 204;
yeccgoto(signalName, 531) ->
 204;
yeccgoto(signalParm, 200) ->
 205;
yeccgoto(signalParm, 207) ->
 210;
yeccgoto(signalParms, 205) ->
 208;
yeccgoto(signalParms, 210) ->
 211;
yeccgoto(signalRequest, 200) ->
 206;
yeccgoto(signalRequest, 207) ->
 206;
yeccgoto(signalRequest, 284) ->
 286;
yeccgoto(signalRequest, 287) ->
 286;
yeccgoto(signalRequest, 521) ->
 527;
yeccgoto(signalRequest, 531) ->
 286;
yeccgoto(signalType, 254) ->
 258;
yeccgoto(signalsDescriptor, 161) ->
 180;
yeccgoto(signalsDescriptor, 182) ->
 180;
yeccgoto(signalsDescriptor, 423) ->
 426;
yeccgoto(signalsDescriptor, 455) ->
 456;
yeccgoto(signalsDescriptor, 790) ->
 809;
yeccgoto(signalsDescriptor, 813) ->
 809;
yeccgoto(signalsDescriptor, 860) ->
 809;
yeccgoto(statisticsDescriptor, 161) ->
 181;
yeccgoto(statisticsDescriptor, 182) ->
 181;
yeccgoto(statisticsDescriptor, 324) ->
 332;
yeccgoto(statisticsDescriptor, 336) ->
 332;
yeccgoto(statisticsDescriptor, 364) ->
 332;
yeccgoto(statisticsDescriptor, 366) ->
 332;
yeccgoto(statisticsDescriptor, 790) ->
 810;
yeccgoto(statisticsDescriptor, 813) ->
 810;
yeccgoto(statisticsDescriptor, 860) ->
 810;
yeccgoto(statisticsParameter, 187) ->
 190;
yeccgoto(statisticsParameter, 191) ->
 194;
yeccgoto(statisticsParameters, 190) ->
 192;
yeccgoto(statisticsParameters, 194) ->
 195;
yeccgoto(streamDescriptor, 324) ->
 333;
yeccgoto(streamDescriptor, 336) ->
 333;
yeccgoto(streamID, 251) ->
 253;
yeccgoto(streamID, 362) ->
 363;
yeccgoto(streamID, 557) ->
 558;
yeccgoto(streamID, 586) ->
 587;
yeccgoto(streamModes, 388) ->
 394;
yeccgoto(streamParm, 324) ->
 334;
yeccgoto(streamParm, 336) ->
 334;
yeccgoto(streamParm, 364) ->
 365;
yeccgoto(streamParm, 366) ->
 369;
yeccgoto(streamParmList, 365) ->
 367;
yeccgoto(streamParmList, 369) ->
 370;
yeccgoto(subtractRequest, 127) ->
 156;
yeccgoto(subtractRequest, 594) ->
 156;
yeccgoto(terminationA, 599) ->
 600;
yeccgoto(terminationA, 603) ->
 600;
yeccgoto(terminationAudit, 790) ->
 811;
yeccgoto(terminationAudit, 860) ->
 861;
yeccgoto(terminationB, 608) ->
 609;
yeccgoto(terminationID, 158) ->
 160;
yeccgoto(terminationID, 295) ->
 297;
yeccgoto(terminationID, 298) ->
 301;
yeccgoto(terminationID, 599) ->
 601;
yeccgoto(terminationID, 603) ->
 601;
yeccgoto(terminationID, 608) ->
 610;
yeccgoto(terminationID, 616) ->
 617;
yeccgoto(terminationID, 622) ->
 623;
yeccgoto(terminationID, 674) ->
 675;
yeccgoto(terminationID, 725) ->
 726;
yeccgoto(terminationID, 728) ->
 729;
yeccgoto(terminationID, 788) ->
 789;
yeccgoto(terminationID, 825) ->
 826;
yeccgoto(terminationID, 850) ->
 851;
yeccgoto(terminationID, 856) ->
 859;
yeccgoto(terminationID, 863) ->
 297;
yeccgoto(terminationID, 876) ->
 859;
yeccgoto(terminationIDList, 293) ->
 296;
yeccgoto(terminationIDList, 857) ->
 865;
yeccgoto(terminationIDList, 877) ->
 865;
yeccgoto(terminationIDListRepeat, 297) ->
 299;
yeccgoto(terminationIDListRepeat, 301) ->
 302;
yeccgoto(terminationStateDescriptor, 324) ->
 335;
yeccgoto(terminationStateDescriptor, 336) ->
 335;
yeccgoto(terminationStateParm, 341) ->
 347;
yeccgoto(terminationStateParm, 348) ->
 351;
yeccgoto(terminationStateParms, 347) ->
 349;
yeccgoto(terminationStateParms, 351) ->
 352;
yeccgoto(timeStamp, 458) ->
 463;
yeccgoto(timeStamp, 479) ->
 463;
yeccgoto(timeStamp, 628) ->
 649;
yeccgoto(timeStamp, 650) ->
 649;
yeccgoto(timeStamp, 684) ->
 463;
yeccgoto(timeStamp, 686) ->
 463;
yeccgoto(timeStamp, 834) ->
 844;
yeccgoto(timeStamp, 845) ->
 844;
yeccgoto(topologyDescriptor, 127) ->
 157;
yeccgoto(topologyDescriptor, 594) ->
 157;
yeccgoto(topologyDescriptor, 765) ->
 157;
yeccgoto(topologyDescriptor, 783) ->
 157;
yeccgoto(topologyDirection, 611) ->
 615;
yeccgoto(topologyTriple, 599) ->
 602;
yeccgoto(topologyTriple, 603) ->
 606;
yeccgoto(topologyTripleList, 602) ->
 604;
yeccgoto(topologyTripleList, 606) ->
 607;
yeccgoto(transactionAck, 741) ->
 743;
yeccgoto(transactionAck, 744) ->
 747;
yeccgoto(transactionAckList, 743) ->
 745;
yeccgoto(transactionAckList, 747) ->
 748;
yeccgoto(transactionID, 115) ->
 733;
yeccgoto(transactionID, 749) ->
 750;
yeccgoto(transactionID, 881) ->
 882;
yeccgoto(transactionItem, 92) ->
 108;
yeccgoto(transactionItem, 108) ->
 108;
yeccgoto(transactionList, 92) ->
 109;
yeccgoto(transactionList, 108) ->
 114;
yeccgoto(transactionPending, 92) ->
 110;
yeccgoto(transactionPending, 108) ->
 110;
yeccgoto(transactionReply, 92) ->
 111;
yeccgoto(transactionReply, 108) ->
 111;
yeccgoto(transactionReplyBody, 753) ->
 757;
yeccgoto(transactionRequest, 92) ->
 112;
yeccgoto(transactionRequest, 108) ->
 112;
yeccgoto(transactionResponseAck, 92) ->
 113;
yeccgoto(transactionResponseAck, 108) ->
 113;
yeccgoto(value, 196) ->
 199;
yeccgoto(value, 227) ->
 238;
yeccgoto(value, 228) ->
 234;
yeccgoto(value, 229) ->
 233;
yeccgoto(value, 230) ->
 232;
yeccgoto(value, 235) ->
 248;
yeccgoto(value, 236) ->
 239;
yeccgoto(value, 240) ->
 246;
yeccgoto(value, 241) ->
 244;
yeccgoto(value, 662) ->
 663;
yeccgoto(valueList, 239) ->
 242;
yeccgoto(valueList, 244) ->
 245;
yeccgoto(valueList, 248) ->
 249;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


